import asyncio
import docker
import functools
import psutil
import time
import uuid
from concurrent.futures import ThreadPoolExecutor
from docker.models.images import Image as DockerImage
from dcrx.image import Image, Label
from dcrx.memory_file import MemoryFile
from dcrx_api.env.time_parser import TimeParser
from dcrx_api.services.registry.models import Registry
from .models import (
    BuildOptions,
    ImageStats,
    JobMetadata
)
from typing import (
    Union, 
    Optional, 
    Dict, 
    Any,
    List
)
from .status import JobStatus
from .connection import JobsConnection


class Job:

    def __init__(
        self, 
        connection: JobsConnection,
        image: Image, 
        registry: Registry,
        build_options: Optional[BuildOptions]=None,
        pool_size: int=psutil.cpu_count(),
        timeout: str='10m',
    ) -> None:

        self.job_id = uuid.uuid4()
        self.job_name = image.full_name

        self.image = image
        self.registry = registry
        self.build_options = build_options

        self.client = docker.DockerClient(
            max_pool_size=pool_size,
        )

        self.context: Union[MemoryFile, None] = None
        self._executor = ThreadPoolExecutor(max_workers=pool_size)
        self.loop = asyncio.get_event_loop()
        self.error: Union[Exception, None] = None
        self.connection = connection

        self.metadata = JobMetadata(
            id=self.job_id,
            image_registry=self.registry.registry_uri,
            name=self.job_name,
            image=self.image.name,
            tag=self.image.tag,
            status=JobStatus.CREATED.value,
            context=f"Job {self.job_id} created."
        )

        dcrx_api_label = Label(
            name='dcrx.api.id',
            value=str(self.job_id)
        )

        self.image.layers.insert(1, dcrx_api_label)
        self.job_pool_size = pool_size
        self.job_image_label = f'{dcrx_api_label.name}={dcrx_api_label.value}'
        
        self.job_start_time = time.monotonic()
        self.timeout = TimeParser(timeout).time

        self.image_stats = ImageStats()
        self.client_closed = False

    async def list(self) -> List[DockerImage]:
        return await self.loop.run_in_executor(
            self._executor,
            functools.partial(
                self.client.images.list,
                self.image.name
            )
        )
    
    async def run(self):
        try:
            await asyncio.wait_for(
                asyncio.create_task(
                    self._run(),
                ),
                timeout=self.timeout
            )
        
        except (asyncio.CancelledError, asyncio.TimeoutError,):

            self.error = Exception(f'Job {self.job_id} timed out')

            self.metadata = JobMetadata(
                id=self.job_id,
                image_registry=self.registry.registry_uri,
                name=self.job_name,
                image=self.image.name,
                tag=self.image.tag,
                status=JobStatus.FAILED.value,
                context=f'Job timed out.'
            )

            await self.connection.update([
                self.metadata
            ], filters={
                'id': self.job_id
            })

            await self.clear()
            await self.close()
  
    async def _run(self):
        await self.connection.create([
            self.metadata
        ])

        try:
            await self.login_to_registry()
            await self.assemble_context()
            await self.build_image()
            await self.push_image()

        except Exception as job_exception:
            self.error = job_exception

        await self.clear()
        await self.close()

    async def login_to_registry(self):

        self.metadata = JobMetadata(
            id=self.job_id,
            image_registry=self.registry.registry_uri,
            name=self.job_name,
            image=self.image.name,
            tag=self.image.tag,
            status=JobStatus.AUTHORIZING.value,
            context=f'Job {self.job_id} authorizing to registry {self.registry.registry_uri}'
        )

        await self.connection.update([
            self.metadata
        ], filters={
            'id': self.job_id
        })

        await self.loop.run_in_executor(
            self._executor,
            functools.partial(
                self.client.api.login,
                self.registry.registry_user,
                password=self.registry.registry_password,
                registry=self.registry.registry_uri,
                reauth=True
            )
        )

    async def assemble_context(self):

        self.metadata = JobMetadata(
            id=self.job_id,
            image_registry=self.registry.registry_uri,
            name=self.job_name,
            image=self.image.name,
            tag=self.image.tag,
            status=JobStatus.GENERATING.value,
            context=f'Job {self.job_id} generating image {self.image.full_name}'
        )

        await self.connection.update([
            self.metadata
        ], filters={
            'id': self.job_id
        })

        self.context = await self.loop.run_in_executor(
            self._executor,
            self.image.to_context
        )


    async def build_image(self):

        self.metadata = JobMetadata(
            id=self.job_id,
            image_registry=self.registry.registry_uri,
            name=self.job_name,
            image=self.image.name,
            tag=self.image.tag,
            status=JobStatus.BUILDING.value,
            context=f'Job {self.job_id} building image {self.image.full_name}'
        )

        await self.connection.update([
            self.metadata
        ], filters={
            'id': self.job_id
        })

        build_options: Dict[str, Any] = {}

        if self.build_options:
            build_options = self.build_options.dict()

        image_result = await self.loop.run_in_executor(
            self._executor,
            functools.partial(
                self.client.images.build,
                dockerfile=self.image.filename,
                fileobj=self.context,
                tag=self.image.full_name,
                custom_context=True,
                forcerm=True,
                **build_options
            )   
        )

        image, _ = image_result 
        self.image_stats = ImageStats(
            size=int(image.attrs.get('Size', 0)/10**6)
        )
        
    async def push_image(self):

        self.metadata = JobMetadata(
            id=self.job_id,
            image_registry=self.registry.registry_uri,
            name=self.job_name,
            image=self.image.name,
            tag=self.image.tag,
            status=JobStatus.PUSHING.value,
            context=f'Job {self.job_id} pushing image {self.image.full_name} to registry {self.registry.registry_uri}',
            size=self.image_stats.size
        )

        await self.connection.update([
            self.metadata
        ], filters={
            'id': self.job_id
        })

        await self.loop.run_in_executor(
            self._executor,
            functools.partial(        
                self.client.images.push,
                self.image.name,
                tag=self.image.tag
            )
        )

    async def clear(self):

        await self.loop.run_in_executor(
            self._executor,
            self.client.api.prune_builds
        )

        await self.loop.run_in_executor(
            self._executor,
            functools.partial(      
                self.client.api.prune_images
            )
        )

        for layer in self.image.layers:
            if layer.layer_type == 'stage':

                try:
                    await self.loop.run_in_executor(
                        self._executor,
                        functools.partial(
                            self.client.images.remove,
                            f'{layer.base}:{layer.tag}'
                        )
                    )
                
                except Exception:
                    pass

        try:
            await self.loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.images.remove,
                    self.image.full_name
                )
            )

        except Exception:
            pass

        await self.loop.run_in_executor(
            self._executor,
            self.image.clear
        )


    async def close(self):

        if self.error is None:
            status = JobStatus.DONE
            context = f'Job {self.job_id} complete'

        elif self.metadata.status == JobStatus.CANCELLED.value:
            status = JobStatus.CANCELLED
            context = f'Job {self.job_id} cancelled'

        else:
            status = JobStatus.FAILED
            context = str(self.error)

        self.metadata = JobMetadata(
            id=self.job_id,
            image_registry=self.registry.registry_uri,
            name=self.job_name,
            image=self.image.name,
            tag=self.image.tag,
            status=status.value,
            context=context,
            size=self.image_stats.size
        )
        

        await self.connection.update([
            self.metadata
        ], filters={
            'id': self.job_id
        })

        if self.client_closed is False:
            await self.loop.run_in_executor(
                None,
                self.client.close
            )
            self.client_closed = True

        if self.context:
            await self.loop.run_in_executor(
                self._executor,
                self.context.close
            )

        
        self._executor.shutdown(cancel_futures=True)
        



    